parser grammar LinkFetchExpParser;
options { tokenVocab=LinkFetchExpLexer; }

/* This will be the entry point of our parser. */
eval
    : directiveExp? labelTagExp? openPageExp? fetchLinkExp afterFetchExp? ';' 
        (alternativeExp labelTagExp? openPageExp? fetchLinkExp afterFetchExp? ';')* 
        propertiesExp?
    ;

directiveExp : DIRECTIVE_START idExp siteExp? paramsExp? C_BRACK ;
idExp : ID_TAG  quotedExp ;
siteExp : SITE_TAG  website (',' website)* ; // for the website expression
website : WEBSITE | quotedExp ;

paramsExp : PARAMS_TAG O_CBR_P paramExp (P_COMMA paramExp)* C_CBR_P ;
paramExp : PARAMETER paramOptionsExp? ;
paramOptionsExp : O_PAREN_P paramOption (pAlt paramOption)* C_PAREN_P ; 
pAlt : P_COMMA | P_OR ;
paramOption : quotedExp ;

// for the open webpage/document expression
//openPageExp : ((labelTagExp? OPEN PAGE usingConnPropsExp?) | labelTagExp) ;
//openPageExp : labelTagExp? onOptionsExp? (OPEN PAGE usingConnPropsExp)? ;
openPageExp : OPEN PAGE usingConnPropsExp? ;

onOptionsExp : ON_TAG optionsValExp (O_COMMA optionsValExp)* C_PAREN_O ;
optionsValExp : optParameter (O_COLON | O_EQ_SYM) optValue ;
optParameter : O_PARAMETER ;
optValue : quotedExp ;

// after fetch expression
afterFetchExp : afterFetchAltExp (',' afterFetchAltExp)* ;
afterFetchAltExp : KEEP_LINKS_REFERRAL_HISTORY | CLOUDFLARE_PROTECTED | ELEVATE_INNER_GRAMMARS ;

// for the alternative command
alternativeExp : OR | AND  ;

// for the fetchlink exp 
fetchLinkExp  
    : onOptionsExp? fetchLinkSubExp (innerAlternativeExp onOptionsExp? fetchLinkSubExp)*
    ;

innerAlternativeExp : OR | ELSE | AND | ALSO ;

fetchLinkSubExp  
    : fetchPageLinksExp 
    | fetchJsonLinksExp 
    | fetchVideoLinksExp 
    | fetchViaCustomHandlerExp  
    | fetchViaUrlRewriteExp
    ;

labelTagExp
    : LABEL_TAG OP_BRACK labelIdExp SEP labelTagValExp (SEP urlRegexExp)? END_LABEL 
    ; // LABEL '[' labelIdExp ',' labelTagValExp (',' urlRegexExp)? ']'

lAlt : EQ_SYM | L_COLON ;
labelIdExp : ID lAlt IDVAL ; // 'id' '=' idval or 'id' ':' idval
labelTagValExp : LABEL lAlt LABELVAL ; // 'label' '=' labelval
urlRegexExp : URL_REGEX lAlt URL_REGEXVAL ; // 'urlregex' = url_regexval

fetchPageLinksExp : (gotoExp)? (waitExp)? (fetchExp | formSubmitExp | webAPICallExp) ;

fetchJsonLinksExp : (waitExp)? fetchJsonExp ;

fetchVideoLinksExp : FETCH VIDEO_LINKS ;

fetchViaCustomHandlerExp 
    : FETCH (fetchQuantifierExp LINKS?)? VIA CUSTOM_HANDLER customHandlerNameExp 
        customHandlerParamsExp? fetchStrategyExp    
;
customHandlerNameExp : quotedExp ;
quotedExp : QUOTED_STR | SINGLE_QUOTED_STR | D_QUOTED_STR | O_QUOTED_STR | P_QUOTED_STR 
          | J_QUOTED_STR | URLSTRING | LABELVAL 
          ;
customHandlerParamsExp 
    : WITH PARAMETERS? '[' customHandlerParamsValExp (',' customHandlerParamsValExp)* ']'
    ;
customHandlerParamsValExp : chParamName EQUAL chParamValue ;
chParamName : quotedExp ;
chParamValue : quotedExp ;

fetchViaUrlRewriteExp : (FETCH | RETURN) REWRITE URL_VERB rewriteExp fetchStrategyExp ;
rewriteExp : REGEXVAL AS? rewriteTerms ;
rewriteTerms : rewriteTerm+ ;
rewriteTerm : URLSCHEME | URLSCHEME_SLASH | SLASH | HOST_VAR | PORT_VAR 
            | PATH_VAR | QUERY_VAR | quotedExp | URLREGEX_GROUP_VAR ;

gotoExp
    : GOTO gotoValExp (',' gotoValExp)*
    ;

gotoValExp 
    : defaultGotoValExp | cssSelectorsGotoValExp
    ;
defaultGotoValExp : '[' ABSOLUTE? htmlTagExp ('->' htmlTagExp)* ']' ;
cssSelectorsGotoValExp : quotedExp ;

htmlTagExp
    : ABS? tagQuantifierExp ':' '<' tagName (tagAttributeExp)* '>'
    ;

tagName : H_WORD ;

tagQuantifierExp : allTagsExp | singleTagExp | rangeOfTagsExp | discreteNumTagsExp ;

allTagsExp : ALL ;
singleTagExp : NUMBER ;
rangeOfTagsExp : NUMBER '-' NUMBER ;
discreteNumTagsExp : NUMBER (',' NUMBER)+ ;

tagAttributeExp
    : attributeName H_ASSIGN attributeValExp           // e.g class = "bold"
    ;

attributeName : H_WORD | H_HTML_ATTR | H_TEXT_ATTR ;  // the last two are special cases $html & $text

attributeValExp
    : ATTRIBVAL | REGEX_ATTRIBVAL
    ;

waitExp : WAIT NUMBER waitUnitsExp EACH? ;
waitUnitsExp : SECS | SECONDS | MINUTES | HOURS ;

/* for the fetch expr */
fetchExp
    : FETCH fetchQuantifierExp linkTargetExp? fetchStrategyExp
    ;

fetchJsonExp : FETCH fetchQuantifierExp? JSONLINKS jsonLinksExp? fetchStrategyExp;
jsonLinksExp : START_JSON_LINKS_MODE jsonLinkPath (J_SEP jsonLinkPath)*  END_JSON_LINKS_MODE ;
jsonLinkPath : jsonRootElemExp (J_ARROW jsonElemExp)* ;
//jsonRootElemExp : J_ABSOLUTE? J_OBJECT (J_O_BRACK jsonQuantifierExp J_C_BRACK)? ; // e.g. absolute $data[1-4]
jsonRootElemExp : J_ABSOLUTE? (
                J_OBJECT (J_O_BRACK jsonQuantifierExp J_C_BRACK)?
                |  jsonQuantifierExp J_COLON J_OBJECT
                ) ; // the commented expr above is ideally what it should be, this expr is to allow more options
                    // so you can say either absolute $data[1-4] or absoulute 1-4:$data
//jsonElemExp : J_ABS? jsonIdentifier (J_O_BRACK jsonQuantifierExp J_C_BRACK)? ; // e.g
jsonElemExp : J_ABS? (
                jsonIdentifier (J_O_BRACK jsonQuantifierExp J_C_BRACK)?
                |  jsonQuantifierExp J_COLON jsonIdentifier 
                ) ; // same explanation as the one above
jsonIdentifier : J_OBJECT | J_ABSOLUTE | J_ABS | J_IDENT | quotedExp | J_REGEXVAL ;
jsonQuantifierExp : jsonQuantifierVal ; //J_O_BRACK jsonQuantifierVal  J_C_BRACK ; // former value
jsonQuantifierVal : allJElems | singleJElem | rangeOfJElems | discreteNumJElems ;
allJElems : J_ALL? ;                        // to allow [] or [all]
singleJElem : J_NUMBER ;                    // [2]
rangeOfJElems : J_NUMBER J_DASH J_NUMBER ;         // [2-5]
discreteNumJElems : J_NUMBER (J_SEP J_NUMBER)+ ;  // [1, 3, 4]


fetchStrategyExp 
    : fetchDirectStrategyExp | fetchIndirectStrategyExp | fetchProgrammaticStrategyExp
    ;

fetchQuantifierExp
    : ALL                               # all
    | FIRST NUMBER                      # firstExpr
    | LAST NUMBER                       # lastExpr
    | FROM NUMBER NEXT NUMBER           # fromNextExpr
    | FIRST? NUMBER YIELDING            # yieldingExpr
    ;

linkTargetExp
    : LINKS                                     //# links
    | textlinksExp                              //# textlinks
    | TAG_LINK_SOURCE (COMMA TAG_LINK_SOURCE)*  //# tagLinkSources
    | SELF_ATTRIBUTE_LINK                       //# selfAttributeLinks
//    | AHREF                             # ahref
//    | IMGSRC                            # imgsrc
    ;

textlinksExp : textlinkTypeExp textlinkPatternExp? ;
textlinkTypeExp : (TEXTLINKS|COMMENTLINKS|CDATALINKS|JSOBFUSCATEDLINKS) ;
textlinkPatternExp : IN PATTERN REGEXVAL ;
//htmlLinkTargetExp  : TAG_LINK_SOURCE ;

fetchDirectStrategyExp
    : DIRECT? usingConnPropsExp? removeExp? customNameExp? customSubFolderExp? nextPageExp?
    ;

fetchIndirectStrategyExp
    : INDIRECT usingConnPropsExp? removeExp? customNameExp? customSubFolderExp? nextPageExp? fetchIndirectExp
    ;

fetchIndirectExp
    : '{' (fetchLinkExp | fetchViaLabelExp) '}'
    ;

fetchViaLabelExp : VIA LABEL_ID NUMBER ;

fetchProgrammaticStrategyExp
    : PROGRAMMATICALLY removeExp?  customNameExp? customSubFolderExp? nextPageExp? 
    ;

usingConnPropsExp 
    : USING CONN_PROPS_STR NUMBER
    ;

removeExp
    : (REMOVE LINKS | ONLY LINKS) MATCHING regexFilterExp
    ;

regexFilterExp : REGEXVAL ;

customNameExp : CUSTOMNAME (VALUE_LITERAL EQUAL)?  customValExp extensionExp? ;
customValExp : PAGE_TITLE_LIT | quotedExp | paramJSVariableExp | paramHtmlValExp | paramRegexExp ;
extensionExp : (WORKOUT FILE? EXTENSION) | (FILE? EXTENSION EQUAL? extensionValExp) ;
extensionValExp : extvalCustomExtExp | extvalFromUrlExp |  WORKOUT | extvalNoneExp | extvalFromContentTypeExp ;
extvalCustomExtExp : quotedExp ;
extvalFromUrlExp : FROM URL_VERB ;
extvalFromContentTypeExp : FROM CONTENT_TYPE ;
extvalNoneExp : NONE | NULL ;

customSubFolderExp : CUSTOMSUBFOLDER (VALUE_LITERAL EQUAL)?  customValExp ;

nextPageExp
    : NEXTPAGE (nextPageLocExp | nextPageCustomHandlerExp) FOLLOW_NEXTPAGES_OF_NEXTPAGES?
    ;
nextPageLocExp : LOCATION gotoValExp TAG_LINK_SOURCE? viaJavascriptExp? ;
nextPageCustomHandlerExp : VIA CUSTOM_HANDLER customHandlerNameExp customHandlerParamsExp? ;

viaJavascriptExp : VIA JAVASCRIPT ;

// for the form submit exp
formSubmitExp 
    : SUBMIT HTMLFORM  toURLExp? formParametersExp? submitMethodExp? usingConnPropsExp? 
      customNameExp? customSubFolderExp? fetchIndirectExp
    ;

toURLExp : TO URLSTRING ;

formParametersExp 
    : WITH PARAMETERS? '[' formParametersValExp (',' formParametersValExp)* ']'
    ;
formParametersValExp : (ALL | formVariable | customParamsExp) ;
formVariable : FORM_VARIABLE | wfKeywordsAsFormVar ; // wf keywords that may overshadow form variable
wfKeywordsAsFormVar : URLSCHEME | HOST_VAR | PORT_VAR | PATH_VAR | QUERY_VAR ; // are put here
customParamsExp : paramNameExp EQUAL paramValExp ;
paramNameExp : quotedExp ;
paramValExp : quotedExp | paramHtmlValExp | paramRegexExp ;
paramRegexExp : NUMBER ':' REGEXVAL ;

submitMethodExp : VIA httpRequestMethodExp ;

// for the api call
webAPICallExp 
    : MAKE API CALL toURLExp apiParametersExp apiRequestMethodExp usingConnPropsExp? 
      customNameExp? customSubFolderExp? fetchIndirectExp
    ;
apiParametersExp 
    : WITH PARAMETERS? '[' apiParametersValExp (',' apiParametersValExp)* ']'
    ;
apiParametersValExp : apiParamNameExp EQUAL apiParamValExp ;
apiParamNameExp : quotedExp ;
apiParamValExp : paramConstantExp | paramJSVariableExp | paramHtmlValExp | paramRegexExp ;
paramConstantExp : quotedExp ;
paramJSVariableExp : JSVAR quotedExp ;
paramHtmlValExp : HTMLVAL GOTO? gotoValExp (SELF_ATTRIBUTE_LINK | SELF_TEXT_VAL | paramRegexExp) ;

apiRequestMethodExp : VIA httpRequestMethodExp ;

httpRequestMethodExp : GET | (POST httpPostEncodingExp?) ;
httpPostEncodingExp : POST_ENCODING quotedExp ;

// for the properties exp
propertiesExp
    : START_CONN_PROP namespaceExp? (propsIDExp)+ END_CONN_PROP
    ;

namespaceExp : NAMESPACE quotedExp ;

propsIDExp 
    : idStartExp 
      requestMethodExp? (headerFieldExp | postParameterExp)* otherPropsExp*
      C_PAREN
    ;

idStartExp
    : ID_START NUM O_PAREN
    ;

requestMethodExp : REQUEST_METHOD REQ_WORD ;

headerFieldExp : HEADER_FIELD keyExp ASSIGN valueExp ;

postParameterExp : POST_PARAMETER keyExp ASSIGN valueExp ;

keyExp : quotedExp ;
valueExp : quotedExp | runtimeVariableExp ;
runtimeVariableExp : VARIABLE | P_VARIABLE ;

otherPropsExp : hotlinkAwareExp | authorizationExp | javascriptReqExp ;

hotlinkAwareExp : HOTLINK_AWARE ;
authorizationExp : AUTHORIZATION_REQUIRED ;
javascriptReqExp : JAVASCRIPT_LOADING_REQUIRED ;
